"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.JsonTemplateParser = void 0;
const error_1 = require("./error");
const types_1 = require("./types");
const MESSAGES = {
    INVALID_ID: 'Invalid variable name "%0"',
    RESERVED_ID: 'Reserved variable name "%0"',
    UNEXP_TOKEN: 'Unexpected token "%0"',
    UNEXP_EOT: 'Unexpected end of template',
};
const EMPTY_EXPR = { type: types_1.SyntaxType.EMPTY };
class JsonTemplateParser {
    constructor(template) {
        this.idx = 0;
        this.codeChars = template.split('');
    }
    static throwUnexpected(token) {
        if (token.type === types_1.TokenType.EOT) {
            this.throwError(token, MESSAGES.UNEXP_EOT);
        }
        this.throwError(token, MESSAGES.UNEXP_TOKEN, token.value);
    }
    static throwError(token, messageFormat, ...args) {
        const msg = messageFormat.replace(/%(\d)/g, (_, idx) => {
            return args[idx] || '';
        });
        throw new error_1.JsonTemplateParserError(msg, token.range[0]);
    }
    validateNoMoreTokensLeft() {
        const token = this.lex();
        if (token.type !== types_1.TokenType.EOT) {
            JsonTemplateParser.throwUnexpected(token);
        }
    }
    static isLiteralToken(token) {
        return (token.type === types_1.TokenType.BOOL ||
            token.type === types_1.TokenType.NUM ||
            token.type === types_1.TokenType.STR ||
            token.type === types_1.TokenType.NULL);
    }
    parse() {
        if (!this.expr) {
            const expr = this.parseStatementsExpr();
            this.validateNoMoreTokensLeft();
            this.expr = expr;
        }
        return this.expr;
    }
    parseStatementsExpr() {
        let statements = [];
        while (!this.match('}')) {
            let expr = this.parseStatementExpr();
            if (expr.type !== types_1.SyntaxType.EMPTY) {
                statements.push(expr);
            }
            if (this.matchTokenType(types_1.TokenType.EOT)) {
                break;
            }
            if (!this.match('}')) {
                this.expect(';');
            }
        }
        return {
            type: types_1.SyntaxType.STATEMENTS_EXPR,
            statements,
        };
    }
    parseStatementExpr() {
        return this.parseLogicalANDExpr();
    }
    parsePathConcatExpr() {
        const expr = this.parsePathConcatPartExpr();
        let operands;
        while (this.match('|')) {
            this.lex();
            (operands || (operands = [expr])).push(this.parsePathConcatPartExpr());
        }
        return operands
            ? {
                type: types_1.SyntaxType.CONCAT_EXPR,
                args: operands,
            }
            : expr;
    }
    parsePathConcatPartExpr() {
        return this.match('(') ? this.parsePathGroupExpr() : this.parsePath();
    }
    parsePathGroupExpr() {
        this.expect('(');
        let expr = this.parseLogicalORExpr();
        this.expect(')');
        let parts = [];
        let part;
        while ((part = this.parsePredicate())) {
            parts.push(part);
        }
        if (!parts.length) {
            return expr;
        }
        else if (expr.type === types_1.SyntaxType.PATH) {
            expr.parts = (expr.parts || []).concat(parts);
            return expr;
        }
        parts.unshift(expr);
        return {
            type: types_1.SyntaxType.PATH,
            parts: parts,
        };
    }
    parsePredicate() {
        if (this.match('[')) {
            return this.parsePosPredicate();
        }
        if (this.match('{')) {
            return this.parseObjectPredicate();
        }
        if (this.match('(')) {
            return this.parsePathGroupExpr();
        }
    }
    parsePath() {
        if (!this.matchPath()) {
            JsonTemplateParser.throwUnexpected(this.lex());
        }
        let root;
        if (this.match('^')) {
            this.lex();
            root = 'data';
        }
        else if (this.matchID()) {
            root = this.lex().value;
        }
        let parts = [];
        let part;
        while ((part = this.parsePathPart())) {
            parts.push(part);
        }
        return {
            type: types_1.SyntaxType.PATH,
            root,
            parts,
        };
    }
    parsePathPart() {
        return this.matchSelector() ? this.parseSelector() : this.parsePredicate();
    }
    parseSelector() {
        let selector = this.lex().value;
        const token = this.lookahead();
        let prop;
        if (this.match('*') || token.type === types_1.TokenType.ID || token.type === types_1.TokenType.STR) {
            prop = this.lex().value;
        }
        return {
            type: types_1.SyntaxType.SELECTOR,
            selector: selector,
            prop,
        };
    }
    parsePosPredicate() {
        this.expect('[');
        let expr = this.parsePosExpr();
        this.expect(']');
        return expr;
    }
    parseObjectPredicate() {
        this.expect('{');
        let expr = this.parseLogicalORExpr();
        this.expect('}');
        return {
            type: types_1.SyntaxType.OBJ_PRED,
            arg: expr,
        };
    }
    parseLogicalORExpr() {
        let expr = this.parseLogicalANDExpr();
        while (this.match('||')) {
            expr = {
                type: types_1.SyntaxType.LOGICAL_EXPR,
                op: this.lex().value,
                args: [expr, this.parseLogicalANDExpr()],
            };
        }
        return expr;
    }
    parseLogicalANDExpr() {
        let expr = this.parseEqualityExpr();
        while (this.match('&&')) {
            expr = {
                type: types_1.SyntaxType.LOGICAL_EXPR,
                op: this.lex().value,
                args: [expr, this.parseEqualityExpr()],
            };
        }
        return expr;
    }
    parseEqualityExpr() {
        let expr = this.parseRelationalExpr();
        while (this.match('==') ||
            this.match('!=') ||
            this.match('===') ||
            this.match('!==') ||
            this.match('^==') ||
            this.match('==^') ||
            this.match('^=') ||
            this.match('=^') ||
            this.match('$==') ||
            this.match('==$') ||
            this.match('$=') ||
            this.match('=$') ||
            this.match('*==') ||
            this.match('==*') ||
            this.match('*=') ||
            this.match('=*')) {
            expr = {
                type: types_1.SyntaxType.COMPARISON_EXPR,
                op: this.lex().value,
                args: [expr, this.parseEqualityExpr()],
            };
        }
        return expr;
    }
    parseRelationalExpr() {
        let expr = this.parseAdditiveExpr();
        while (this.match('<') || this.match('>') || this.match('<=') || this.match('>=')) {
            expr = {
                type: types_1.SyntaxType.COMPARISON_EXPR,
                op: this.lex().value,
                args: [expr, this.parseRelationalExpr()],
            };
        }
        return expr;
    }
    parseAdditiveExpr() {
        let expr = this.parseMultiplicativeExpr();
        while (this.match('+') || this.match('-')) {
            expr = {
                type: types_1.SyntaxType.MATH_EXPR,
                op: this.lex().value,
                args: [expr, this.parseMultiplicativeExpr()],
            };
        }
        return expr;
    }
    parseMultiplicativeExpr() {
        let expr = this.parseUnaryExpr();
        while (this.match('*') || this.match('/') || this.match('%')) {
            expr = {
                type: types_1.SyntaxType.MATH_EXPR,
                op: this.lex().value,
                args: [expr, this.parseMultiplicativeExpr()],
            };
        }
        return expr;
    }
    parsePosExpr() {
        if (this.match(':')) {
            this.lex();
            return {
                type: types_1.SyntaxType.POS_EXPR,
                toIdx: this.parseLogicalORExpr(),
            };
        }
        let fromExpr = this.parseLogicalORExpr();
        if (this.match(':')) {
            this.lex();
            if (this.match(']')) {
                return {
                    type: types_1.SyntaxType.POS_EXPR,
                    fromIdx: fromExpr,
                };
            }
            return {
                type: types_1.SyntaxType.POS_EXPR,
                fromIdx: fromExpr,
                toIdx: this.parseLogicalORExpr(),
            };
        }
        return {
            type: types_1.SyntaxType.POS_EXPR,
            idx: fromExpr,
        };
    }
    parseUnaryExpr() {
        if (this.match('!') || this.match('-') || this.match('...')) {
            return {
                type: types_1.SyntaxType.UNARY_EXPR,
                op: this.lex().value,
                arg: this.parseUnaryExpr(),
            };
        }
        return this.parsePrimaryExpr();
    }
    parseLiteralExpr() {
        return {
            type: types_1.SyntaxType.LITERAL,
            value: this.lex().value,
        };
    }
    parsePathVariable() {
        const idParts = [];
        while (this.matchID()) {
            idParts.push(this.lex().value);
            if (this.match('.')) {
                this.lex();
            }
        }
        if (!idParts.length) {
            JsonTemplateParser.throwUnexpected(this.lex());
        }
        return idParts.join('.');
    }
    parseAssignmentExpr() {
        this.expectTokenType(types_1.TokenType.ASSIGNMENT);
        const id = this.parsePathVariable();
        this.expect('=');
        return {
            type: types_1.SyntaxType.ASSIGNMENT_EXPR,
            id,
            value: this.parseLogicalORExpr(),
        };
    }
    parseFunctionCallExpr() {
        this.expectTokenType(types_1.TokenType.FUNCTION_CALL);
        const id = this.parsePathVariable();
        const args = [];
        this.expect('(');
        while (!this.match(')')) {
            args.push(this.parseLogicalORExpr());
            if (!this.match(')')) {
                this.expect(',');
            }
        }
        this.expect(')');
        return {
            type: types_1.SyntaxType.FUNCTION_CALL_EXPR,
            id,
            args,
        };
    }
    parsePrimaryExpr() {
        if (this.matchEOT() || this.match(';')) {
            return EMPTY_EXPR;
        }
        let token = this.lookahead();
        if (token.type === types_1.TokenType.ASSIGNMENT) {
            return this.parseAssignmentExpr();
        }
        if (token.type === types_1.TokenType.FUNCTIION) {
            return this.parseFunctionExpr();
        }
        if (token.type === types_1.TokenType.FUNCTION_CALL) {
            return this.parseFunctionCallExpr();
        }
        if (JsonTemplateParser.isLiteralToken(token)) {
            return this.parseLiteralExpr();
        }
        if (this.matchSelector()) {
            return this.parsePathConcatExpr();
        }
        if (this.matchPath()) {
            return this.parsePath();
        }
        if (this.match('(')) {
            return this.parseGroupExpr();
        }
        if (this.match('{')) {
            return this.parseObjectExpr();
        }
        if (this.match('[')) {
            return this.parseArrayExpr();
        }
        return JsonTemplateParser.throwUnexpected(this.lex());
    }
    parseFunctionExpr() {
        this.expectTokenType(types_1.TokenType.FUNCTIION);
        this.expect('(');
        const params = [];
        while (!this.match(')')) {
            const token = this.lookahead();
            if (token.type !== types_1.TokenType.ID) {
                JsonTemplateParser.throwUnexpected(token);
            }
            params.push(this.lex().value);
            if (!this.match(')')) {
                this.expect(',');
            }
        }
        this.expect(')');
        this.expect('{');
        const body = this.parseStatementsExpr();
        this.expect('}');
        return {
            type: types_1.SyntaxType.FUNCTION_EXPR,
            params,
            body,
        };
    }
    parseObjectExpr() {
        this.expect('{');
        const props = {};
        while (!this.match('}')) {
            const token = this.lookahead();
            if (token.type !== types_1.TokenType.STR && token.type !== types_1.TokenType.ID) {
                JsonTemplateParser.throwUnexpected(token);
            }
            const name = this.lex().value;
            this.expect(':');
            const value = this.parseLogicalANDExpr();
            props[name] = value;
            if (!this.match('}')) {
                this.expect(',');
            }
        }
        this.expect('}');
        return { type: types_1.SyntaxType.OBJECT_EXPR, props };
    }
    parseArrayExpr() {
        this.expect('[');
        const elements = [];
        while (!this.match(']')) {
            const expr = this.parseLogicalANDExpr();
            elements.push(expr);
            if (!this.match(']')) {
                this.expect(',');
            }
        }
        this.expect(']');
        return { type: types_1.SyntaxType.ARRAY_EXPR, elements };
    }
    parseGroupExpr() {
        this.expect('(');
        let expr = this.parseLogicalORExpr();
        this.expect(')');
        return expr;
    }
    match(value) {
        let token = this.lookahead();
        return token.type === types_1.TokenType.PUNCT && token.value === value;
    }
    matchPath() {
        return this.matchSelector() || this.matchID() || this.match('^');
    }
    matchSelector() {
        let token = this.lookahead();
        if (token.type === types_1.TokenType.PUNCT) {
            let value = token.value;
            return value === '.' || value === '..';
        }
        return false;
    }
    matchTokenType(tokenType) {
        let token = this.lookahead();
        return token.type === tokenType;
    }
    matchToken(matchToken) {
        let token = this.lookahead();
        return token.type === matchToken.type && token.value === matchToken.value;
    }
    matchID() {
        return this.matchTokenType(types_1.TokenType.ID);
    }
    matchEOT() {
        return this.matchTokenType(types_1.TokenType.EOT);
    }
    expectTokenType(tokenType) {
        let token = this.lex();
        if (token.type !== tokenType) {
            JsonTemplateParser.throwUnexpected(token);
        }
    }
    expect(value) {
        let token = this.lex();
        if (token.type !== types_1.TokenType.PUNCT || token.value !== value) {
            JsonTemplateParser.throwUnexpected(token);
        }
    }
    lookahead() {
        if (this.buf !== undefined) {
            return this.buf;
        }
        let pos = this.idx;
        this.buf = this.advance();
        this.idx = pos;
        return this.buf;
    }
    advance() {
        while (JsonTemplateParser.isWhiteSpace(this.codeChars[this.idx])) {
            ++this.idx;
        }
        if (this.idx >= this.codeChars.length) {
            return {
                type: types_1.TokenType.EOT,
                range: [this.idx, this.idx],
            };
        }
        let token = this.scanPunctuator() || this.scanID() || this.scanString() || this.scanNumeric();
        if (token) {
            return token;
        }
        // Invalid Token
        token = { range: [this.idx, this.idx], type: types_1.TokenType.UNKNOWN };
        if (this.idx >= this.codeChars.length) {
            token.type = types_1.TokenType.EOT;
        }
        else {
            token.value = this.codeChars[this.idx];
        }
        JsonTemplateParser.throwUnexpected(token);
    }
    lex() {
        let token;
        if (this.buf) {
            this.idx = this.buf.range[1];
            token = this.buf;
            this.buf = undefined;
            return token;
        }
        return this.advance();
    }
    static isDigit(ch) {
        return '0123456789'.indexOf(ch) >= 0;
    }
    static isWhiteSpace(ch) {
        return ' \r\n\t'.indexOf(ch) > -1;
    }
    static isIdStart(ch) {
        return (ch === '$' || ch === '@' || ch === '_' || (ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z'));
    }
    static isIdPart(ch) {
        return this.isIdStart(ch) || (ch >= '0' && ch <= '9');
    }
    static validateIDToken(token) {
        if (/^__\d+(\.|$)/.test(token.value)) {
            this.throwError(token, MESSAGES.RESERVED_ID, token.value);
        }
        if (token.value.endsWith('.')) {
            this.throwError(token, MESSAGES.INVALID_ID, token.value);
        }
    }
    scanID() {
        let ch = this.codeChars[this.idx];
        if (!JsonTemplateParser.isIdStart(ch)) {
            return;
        }
        let start = this.idx, id = ch;
        while (++this.idx < this.codeChars.length) {
            ch = this.codeChars[this.idx];
            if (!JsonTemplateParser.isIdPart(ch)) {
                break;
            }
            id += ch;
        }
        switch (id) {
            case 'true':
            case 'false':
                return {
                    type: types_1.TokenType.BOOL,
                    value: id === 'true',
                    range: [start, this.idx],
                };
            case 'null':
                return {
                    type: types_1.TokenType.NULL,
                    value: null,
                    range: [start, this.idx],
                };
            case 'set':
                return {
                    type: types_1.TokenType.ASSIGNMENT,
                    range: [start, this.idx],
                };
            case 'function':
                return {
                    type: types_1.TokenType.FUNCTIION,
                    value: 'function',
                    range: [start, this.idx],
                };
            case 'call':
                return {
                    type: types_1.TokenType.FUNCTION_CALL,
                    value: 'call',
                    range: [start, this.idx],
                };
            default:
                const token = {
                    type: types_1.TokenType.ID,
                    value: id.replace(/^\$/, 'bindings.'),
                    range: [start, this.idx],
                };
                JsonTemplateParser.validateIDToken(token);
                return token;
        }
    }
    scanString() {
        if (this.codeChars[this.idx] !== '"' && this.codeChars[this.idx] !== "'") {
            return;
        }
        let orig = this.codeChars[this.idx], start = ++this.idx, str = '', eosFound = false, ch;
        while (this.idx < this.codeChars.length) {
            ch = this.codeChars[this.idx++];
            if (ch === '\\') {
                ch = this.codeChars[this.idx++];
            }
            else if ((ch === '"' || ch === "'") && ch === orig) {
                eosFound = true;
                break;
            }
            str += ch;
        }
        if (eosFound) {
            return {
                type: types_1.TokenType.STR,
                value: str,
                range: [start, this.idx],
            };
        }
    }
    scanNumeric() {
        let start = this.idx, ch = this.codeChars[this.idx], isFloat = ch === '.';
        if (isFloat || JsonTemplateParser.isDigit(ch)) {
            let num = ch;
            while (++this.idx < this.codeChars.length) {
                ch = this.codeChars[this.idx];
                if (ch === '.') {
                    if (isFloat) {
                        return;
                    }
                    isFloat = true;
                }
                else if (!JsonTemplateParser.isDigit(ch)) {
                    break;
                }
                num += ch;
            }
            return {
                type: types_1.TokenType.NUM,
                value: isFloat ? parseFloat(num) : parseInt(num, 10),
                range: [start, this.idx],
            };
        }
    }
    scanPunctuatorForDots() {
        let start = this.idx, ch1 = this.codeChars[this.idx], ch2 = this.codeChars[this.idx + 1], ch3 = this.codeChars[this.idx + 2];
        if (ch1 !== '.') {
            return;
        }
        if (ch2 === '.' && ch3 === '.') {
            this.idx = this.idx + 3;
            return {
                type: types_1.TokenType.PUNCT,
                value: '...',
                range: [start, this.idx],
            };
        }
        else if (ch2 === '.') {
            this.idx = this.idx + 2;
            return {
                type: types_1.TokenType.PUNCT,
                value: '..',
                range: [start, this.idx],
            };
        }
        else {
            if (JsonTemplateParser.isDigit(ch2)) {
                return;
            }
            return {
                type: types_1.TokenType.PUNCT,
                value: '.',
                range: [start, ++this.idx],
            };
        }
    }
    scanPunctuatorForEquality() {
        let start = this.idx, ch1 = this.codeChars[this.idx], ch2 = this.codeChars[this.idx + 1], ch3 = this.codeChars[this.idx + 2];
        if (ch2 === '=') {
            if (ch3 === '=') {
                if ('=!^$*'.indexOf(ch1) >= 0) {
                    this.idx += 3;
                    return {
                        type: types_1.TokenType.PUNCT,
                        value: ch1 + ch2 + ch3,
                        range: [start, this.idx],
                    };
                }
            }
            else if ('^$*'.indexOf(ch3) >= 0) {
                if (ch1 === '=') {
                    this.idx += 3;
                    return {
                        type: types_1.TokenType.PUNCT,
                        value: ch1 + ch2 + ch3,
                        range: [start, this.idx],
                    };
                }
            }
            else if ('=!^$*><'.indexOf(ch1) >= 0) {
                this.idx += 2;
                return {
                    type: types_1.TokenType.PUNCT,
                    value: ch1 + ch2,
                    range: [start, this.idx],
                };
            }
        }
        else if (ch1 === '=') {
            if ('^$*'.indexOf(ch2) >= 0) {
                this.idx += 2;
                return {
                    type: types_1.TokenType.PUNCT,
                    value: ch1 + ch2,
                    range: [start, this.idx],
                };
            }
            else {
                return {
                    type: types_1.TokenType.PUNCT,
                    value: ch1,
                    range: [start, ++this.idx],
                };
            }
        }
    }
    scanPunctuatorForLogicalTokens() {
        let start = this.idx, ch1 = this.codeChars[this.idx], ch2 = this.codeChars[this.idx + 1];
        if (ch1 === ch2 && (ch1 === '|' || ch1 === '&')) {
            this.idx += 2;
            return {
                type: types_1.TokenType.PUNCT,
                value: ch1 + ch2,
                range: [start, this.idx],
            };
        }
    }
    scanSingleCharPunctuators() {
        let start = this.idx, ch1 = this.codeChars[this.idx];
        if (',;:{}()[]^+-*/%!><|'.indexOf(ch1) >= 0) {
            return {
                type: types_1.TokenType.PUNCT,
                value: ch1,
                range: [start, ++this.idx],
            };
        }
    }
    scanPunctuator() {
        return (this.scanPunctuatorForDots() ||
            this.scanPunctuatorForLogicalTokens() ||
            this.scanPunctuatorForEquality() ||
            this.scanSingleCharPunctuators());
    }
}
exports.JsonTemplateParser = JsonTemplateParser;
